"""
MobileNet v4, adapted from
https://github.com/jaiwei98/MobileNetV4-pytorch/blob/main/mobilenet/mobilenetv4.py
and
https://github.com/huggingface/pytorch-image-models/blob/main/timm/layers/attention2d.py

Paper "MobileNetV4 -- Universal Models for the Mobile Ecosystem", https://arxiv.org/abs/2404.10518
"""

# Reference license: MIT and Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerScale2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.mobilenet_v4 import ConvNormActConfig
from birder.net.mobilenet_v4 import InvertedResidual
from birder.net.mobilenet_v4 import InvertedResidualConfig
from birder.net.mobilenet_v4 import UniversalInvertedBottleneck
from birder.net.mobilenet_v4 import UniversalInvertedBottleneckConfig


class MultiQueryAttentionBlockConfig:
    def __init__(
        self,
        in_channels: int,
        dw_kernel_size: tuple[int, int],
        num_heads: int,
        key_dim: int,
        value_dim: int,
        kv_strides: tuple[int, int],
        query_strides: tuple[int, int],
        dropout: float,
    ) -> None:
        self.in_channels = in_channels
        self.dw_kernel_size = dw_kernel_size
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.value_dim = value_dim
        self.kv_strides = kv_strides
        self.query_strides = query_strides
        self.dropout = dropout

        # Compatibility
        self.stride = self.query_strides
        self.out_channels = self.in_channels


class MultiQueryAttention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_heads: int,
        key_dim: int,
        value_dim: int,
        query_strides: tuple[int, int],
        kv_strides: tuple[int, int],
        dw_kernel_size: tuple[int, int],
        dropout: float,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.key_dim = key_dim
        self.query_strides = query_strides

        query_layers = []
        if query_strides[0] > 1 or query_strides[0] > 1:
            query_layers.append(nn.AvgPool2d(kernel_size=query_strides, stride=query_strides, padding=(0, 0)))
            query_layers.append(nn.BatchNorm2d(in_channels))

        query_layers.append(
            nn.Conv2d(
                in_channels,
                self.num_heads * self.key_dim,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
            )
        )
        self.query = nn.Sequential(*query_layers)

        key_layers = []
        value_layers = []
        if kv_strides[0] > 1 or kv_strides[1] > 1:
            key_layers.append(
                Conv2dNormActivation(
                    in_channels,
                    in_channels,
                    kernel_size=dw_kernel_size,
                    stride=kv_strides,
                    padding=((dw_kernel_size[0] - 1) // 2, (dw_kernel_size[1] - 1) // 2),
                    groups=in_channels,
                    activation_layer=None,
                )
            )
            value_layers.append(
                Conv2dNormActivation(
                    in_channels,
                    in_channels,
                    kernel_size=dw_kernel_size,
                    stride=kv_strides,
                    padding=((dw_kernel_size[0] - 1) // 2, (dw_kernel_size[1] - 1) // 2),
                    groups=in_channels,
                    activation_layer=None,
                )
            )

        key_layers.append(
            nn.Conv2d(in_channels, self.key_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        )
        value_layers.append(
            nn.Conv2d(in_channels, value_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False)
        )
        self.key = nn.Sequential(*key_layers)
        self.value = nn.Sequential(*value_layers)

        output_layers = []
        if query_strides[0] > 1 or query_strides[0] > 1:
            output_layers.append(nn.Upsample(scale_factor=self.query_strides, mode="bilinear", align_corners=False))

        output_layers.append(
            nn.Conv2d(
                self.num_heads * value_dim, in_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
            )
        )
        output_layers.append(nn.Dropout(p=dropout))
        self.output = nn.Sequential(*output_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, C, H, W) = x.size()
        q = self.query(x)
        q = q.reshape(B, self.num_heads, self.key_dim, -1)
        q = q.transpose(-1, -2).contiguous()

        k = self.key(x)
        (B, C, _, _) = k.size()
        k = k.reshape(B, C, -1).transpose(1, 2)
        k = k.unsqueeze(1).contiguous()

        v = self.value(x)
        (B, C, _, _) = v.size()
        v = v.reshape(B, C, -1).transpose(1, 2)
        v = v.unsqueeze(1).contiguous()

        # Calculate attention score
        attn_score = F.scaled_dot_product_attention(q, k, v, dropout_p=0.0)  # pylint: disable=not-callable
        (B, _, _, C) = attn_score.size()
        feat_dim = C * self.num_heads
        attn_score = attn_score.transpose(1, 2)
        attn_score = (
            attn_score.reshape(B, H // self.query_strides[0], W // self.query_strides[1], feat_dim)
            .permute(0, 3, 1, 2)
            .contiguous()
        )
        x = self.output(attn_score)

        return x


class MultiQueryAttentionBlock(nn.Module):
    def __init__(self, cnf: MultiQueryAttentionBlockConfig, stochastic_depth_prob: float) -> None:
        super().__init__()

        self.input_norm = nn.BatchNorm2d(cnf.in_channels)
        self.multi_query_attention = MultiQueryAttention(
            cnf.in_channels,
            cnf.num_heads,
            cnf.key_dim,
            cnf.value_dim,
            cnf.query_strides,
            cnf.kv_strides,
            cnf.dw_kernel_size,
            cnf.dropout,
        )
        self.layer_scale = LayerScale2d(cnf.in_channels, 1e-5)
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.input_norm(x)
        x = self.multi_query_attention(x)
        x = self.layer_scale(x)
        x = self.stochastic_depth(x) + shortcut

        return x


# pylint: disable=invalid-name
class MobileNet_v4_Hybrid(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-branches
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        dropout: float = self.config["dropout"]
        drop_path_rate: float = self.config["drop_path_rate"]
        net_size: str = self.config["net_size"]

        if net_size == "medium":
            stem_settings = ConvNormActConfig(0, 32, (3, 3), (2, 2), (1, 1))
            net_settings: list[Any] = [
                # Stage 1
                InvertedResidualConfig(32, 48, (3, 3), (2, 2), (1, 1), 4.0, False),
                # Stage 2
                UniversalInvertedBottleneckConfig(48, 80, 4.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(80, 80, 2.0, (3, 3), (3, 3), (1, 1), True),
                # Stage 3
                UniversalInvertedBottleneckConfig(80, 160, 6.0, (3, 3), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(160, 160, 2.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (5, 5), (1, 1), True),
                MultiQueryAttentionBlockConfig(160, (3, 3), 4, 64, 64, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                MultiQueryAttentionBlockConfig(160, (3, 3), 4, 64, 64, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), None, (1, 1), True),
                MultiQueryAttentionBlockConfig(160, (3, 3), 4, 64, 64, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), (3, 3), (1, 1), True),
                MultiQueryAttentionBlockConfig(160, (3, 3), 4, 64, 64, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(160, 160, 4.0, (3, 3), None, (1, 1), True),
                # Stage 4
                UniversalInvertedBottleneckConfig(160, 256, 6.0, (5, 5), (5, 5), (2, 2), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 2.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 2.0, (3, 3), (5, 5), (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 2.0, None, None, (1, 1), True),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, None, None, (1, 1), True),
                MultiQueryAttentionBlockConfig(256, (3, 3), 4, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (3, 3), None, (1, 1), True),
                MultiQueryAttentionBlockConfig(256, (3, 3), 4, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), (5, 5), (1, 1), True),
                MultiQueryAttentionBlockConfig(256, (3, 3), 4, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), None, (1, 1), True),
                MultiQueryAttentionBlockConfig(256, (3, 3), 4, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(256, 256, 4.0, (5, 5), None, (1, 1), True),
            ]
            last_stage_settings = [
                ConvNormActConfig(256, 960, (1, 1), (1, 1), (0, 0)),
            ]
            features_stage_settings = ConvNormActConfig(960, 1280, (1, 1), (1, 1), (0, 0))
        elif net_size == "large":
            stem_settings = ConvNormActConfig(0, 24, (3, 3), (2, 2), (1, 1), nn.GELU)
            net_settings = [
                # Stage 1
                InvertedResidualConfig(24, 48, (3, 3), (2, 2), (1, 1), 4.0, False, nn.GELU),
                # Stage 2
                UniversalInvertedBottleneckConfig(48, 96, 4.0, (3, 3), (5, 5), (2, 2), True, nn.GELU),
                UniversalInvertedBottleneckConfig(96, 96, 4.0, (3, 3), (3, 3), (1, 1), True, nn.GELU),
                # Stage 3
                UniversalInvertedBottleneckConfig(96, 192, 4.0, (3, 3), (5, 5), (2, 2), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), (5, 5), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(192, (3, 3), 8, 48, 48, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(192, (3, 3), 8, 48, 48, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(192, (3, 3), 8, 48, 48, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(192, (3, 3), 8, 48, 48, (2, 2), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(192, 192, 4.0, (3, 3), None, (1, 1), True, nn.GELU),
                # Stage 4
                UniversalInvertedBottleneckConfig(192, 512, 4.0, (5, 5), (5, 5), (2, 2), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (3, 3), (1, 1), True, nn.GELU),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), (5, 5), (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(512, (3, 3), 8, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(512, (3, 3), 8, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(512, (3, 3), 8, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
                MultiQueryAttentionBlockConfig(512, (3, 3), 8, 64, 64, (1, 1), (1, 1), 0.0),
                UniversalInvertedBottleneckConfig(512, 512, 4.0, (5, 5), None, (1, 1), True, nn.GELU),
            ]
            last_stage_settings = [
                ConvNormActConfig(512, 960, (1, 1), (1, 1), (0, 0), nn.GELU),
            ]
            features_stage_settings = ConvNormActConfig(960, 1280, (1, 1), (1, 1), (0, 0), nn.GELU)
        else:
            raise ValueError(f"Unknown net_size '{net_size}'")

        self.stem = Conv2dNormActivation(
            self.input_channels,
            stem_settings.out_channels,
            kernel_size=stem_settings.kernel,
            stride=stem_settings.stride,
            padding=stem_settings.padding,
            activation_layer=stem_settings.activation,
            inplace=None,
        )

        layers: list[nn.Module] = []
        total_stage_blocks = len(net_settings) + len(last_stage_settings)
        i = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx, block_settings in enumerate(net_settings):
            # Adjust stochastic depth probability based on the depth of the stage block
            sd_prob = drop_path_rate * float(idx) / total_stage_blocks

            if idx > 0 and (block_settings.stride[0] > 1 or block_settings.stride[1] > 1):
                stages[f"stage{i+1}"] = nn.Sequential(*layers)
                return_channels.append(net_settings[idx - 1].out_channels)
                layers = []
                i += 1

            if isinstance(block_settings, ConvNormActConfig):
                layers.append(
                    Conv2dNormActivation(
                        block_settings.in_channels,
                        block_settings.out_channels,
                        kernel_size=block_settings.kernel,
                        stride=block_settings.stride,
                        padding=block_settings.padding,
                        activation_layer=block_settings.activation,
                        inplace=None,
                    )
                )

            elif isinstance(block_settings, InvertedResidualConfig):
                layers.append(InvertedResidual(block_settings, sd_prob))

            elif isinstance(block_settings, UniversalInvertedBottleneckConfig):
                layers.append(UniversalInvertedBottleneck(block_settings, 1e-5, sd_prob))

            elif isinstance(block_settings, MultiQueryAttentionBlockConfig):
                layers.append(MultiQueryAttentionBlock(block_settings, sd_prob))

            else:
                raise ValueError("Unknown config")

        stages[f"stage{i+1}"] = nn.Sequential(*layers)
        return_channels.append(net_settings[-1].out_channels)
        layers = []
        i += 1
        for block_settings in last_stage_settings:
            layers.append(
                Conv2dNormActivation(
                    block_settings.in_channels,
                    block_settings.out_channels,
                    kernel_size=block_settings.kernel,
                    stride=block_settings.stride,
                    padding=block_settings.padding,
                    activation_layer=block_settings.activation,
                    inplace=None,
                )
            )
        stages[f"stage{i+1}"] = nn.Sequential(*layers)
        return_channels.append(last_stage_settings[-1].out_channels)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            Conv2dNormActivation(
                features_stage_settings.in_channels,
                features_stage_settings.out_channels,
                kernel_size=features_stage_settings.kernel,
                stride=features_stage_settings.stride,
                padding=features_stage_settings.padding,
                activation_layer=features_stage_settings.activation,
                inplace=None,
            ),
            nn.Flatten(1),
            nn.Dropout(p=dropout),
        )
        self.return_channels = return_channels[:4]
        self.embedding_size = features_stage_settings.out_channels
        self.classifier = self.create_classifier()

        self.stem_stride = stem_settings.stride[0]
        self.stem_width = stem_settings.out_channels
        self.encoding_size = features_stage_settings.out_channels

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def transform_to_backbone(self) -> None:
        self.body.stage5 = nn.Identity()
        self.features = nn.Identity()
        self.classifier = nn.Identity()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config(
    "mobilenet_v4_hybrid_m", MobileNet_v4_Hybrid, config={"dropout": 0.2, "drop_path_rate": 0.075, "net_size": "medium"}
)
registry.register_model_config(
    "mobilenet_v4_hybrid_l", MobileNet_v4_Hybrid, config={"dropout": 0.2, "drop_path_rate": 0.35, "net_size": "large"}
)

registry.register_weights(
    "mobilenet_v4_hybrid_m_il-common",
    {
        "description": "MobileNet hybrid v4 medium model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 39.7,
                "sha256": "220df49e08ea49e24f30dcc777bf48c7aaea4aaa5909b56c931f41747381d390",
            }
        },
        "net": {"network": "mobilenet_v4_hybrid_m", "tag": "il-common"},
    },
)



"""
MAE Hiera, adapted from
https://github.com/facebookresearch/hiera/blob/main/hiera/hiera_mae.py

Paper "Hiera: A Hierarchical Vision Transformer without the Bells-and-Whistles", https://arxiv.org/abs/2306.00989
"""

# Reference license: Apache-2.0

import math
from typing import Any
from typing import Optional

import torch
from torch import nn

from birder.common.masking import uniform_mask
from birder.model_registry import registry
from birder.net.hiera import Hiera
from birder.net.hiera import HieraBlock
from birder.net.hiera import undo_windowing
from birder.net.mim.base import MIMBaseNet


def apply_fusion_head(head: nn.Module, x: torch.Tensor) -> torch.Tensor:
    if isinstance(head, nn.Identity):
        return x

    (B, num_mask_units) = x.shape[0:2]

    # Apply head, e.g [B, #MUs, My, Mx, C] -> head([B * #MUs, C, My, Mx])
    permute = [0] + [len(x.shape) - 2] + list(range(1, len(x.shape) - 2))
    x = head(x.reshape(B * num_mask_units, *x.shape[2:]).permute(permute))

    # Restore original layout, e.g. [B * #MUs, C', My', Mx'] -> [B, #MUs, My', Mx', C']
    permute = [0] + list(range(2, len(x.shape))) + [1]
    x = x.permute(permute).reshape(B, num_mask_units, *x.shape[2:], x.shape[1])

    return x


# pylint: disable=invalid-name
class MAE_Hiera(MIMBaseNet):
    default_size = (224, 224)

    def __init__(
        self,
        encoder: Hiera,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(encoder, config=config, size=size)
        assert self.config is None, "config not supported"
        assert isinstance(self.encoder, Hiera)

        self.mask_ratio = 0.6
        encoder_dim = self.encoder.encoding_size
        decoder_embed_dim = 512
        decoder_depth = 8
        decoder_num_heads = 16
        self.patch_size = self.encoder.mask_unit_size[0] * self.encoder.patch_stride[0]

        self.mask_unit_spatial_shape_final = [
            i // s ** (self.encoder.q_pool) for i, s in zip(self.encoder.mask_unit_size, self.encoder.q_stride)
        ]
        self.tokens_spatial_shape_final = [
            i // s ** (self.encoder.q_pool) for i, s in zip(self.encoder.tokens_spatial_shape, self.encoder.q_stride)
        ]

        self.encoder_norm = nn.LayerNorm(encoder_dim)

        curr_mu_size = list(self.encoder.mask_unit_size)
        self.multi_scale_fusion_heads = nn.ModuleList()
        stage_idx = 0
        for _ in self.encoder.stage_ends[: self.encoder.q_pool]:  # resolution constant after q_pool
            kernel = [i // s for i, s in zip(curr_mu_size, self.mask_unit_spatial_shape_final)]
            curr_mu_size = [i // s for i, s in zip(curr_mu_size, self.encoder.q_stride)]
            self.multi_scale_fusion_heads.append(
                nn.Conv2d(
                    self.encoder.return_channels[stage_idx],
                    encoder_dim,
                    kernel_size=kernel,
                    stride=kernel,
                    padding=(0, 0),
                )
            )
            stage_idx += 1

        self.multi_scale_fusion_heads.append(nn.Identity())  # final stage, no transform

        # MAE decoder
        self.decoder_embed = nn.Linear(encoder_dim, decoder_embed_dim)
        self.mask_token = nn.Parameter(torch.zeros(1, 1, decoder_embed_dim))

        self.decoder_pos_embed = nn.Parameter(
            torch.zeros(1, math.prod(self.tokens_spatial_shape_final), decoder_embed_dim)
        )

        layers = []
        for _ in range(decoder_depth):
            layers.append(
                HieraBlock(
                    dim=decoder_embed_dim,
                    dim_out=decoder_embed_dim,
                    heads=decoder_num_heads,
                    mlp_ratio=4.0,
                    drop_path=0.0,
                    q_stride=1,
                    window_size=0,
                    use_mask_unit_attn=False,
                )
            )

        self.decoder_blocks = nn.Sequential(*layers)
        self.decoder_norm = nn.LayerNorm(decoder_embed_dim)

        # Predictor
        self.pred_stride = self.encoder.patch_stride[-1] * (self.encoder.q_stride[-1] ** self.encoder.q_pool)

        self.decoder_pred = nn.Linear(
            decoder_embed_dim,
            (self.pred_stride ** min(2, len(self.encoder.q_stride))) * self.encoder.input_channels,
        )

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.weight, 1.0)
                nn.init.zeros_(m.bias)

    def get_pixel_label_2d(self, input_img: torch.Tensor, mask: torch.Tensor, norm: bool) -> torch.Tensor:
        input_img = input_img.permute(0, 2, 3, 1)

        size = self.pred_stride
        label = input_img.unfold(1, size, size).unfold(2, size, size)
        label = label.flatten(1, 2).flatten(2)
        label = label[mask]
        if norm is True:
            mean = label.mean(dim=-1, keepdim=True)
            var = label.var(dim=-1, keepdim=True)
            label = (label - mean) / (var + 1.0e-6) ** 0.5

        return label

    def unpatchify(self, x: torch.Tensor) -> torch.Tensor:
        """
        x: (N, L, patch_size**2 * 3)
        imgs: (N, 3, H, W)
        """

        p = self.pred_stride
        h = int(x.shape[1] ** 0.5)
        w = int(x.shape[1] ** 0.5)
        assert h * w == x.shape[1]

        x = x.reshape(shape=(x.shape[0], h, w, p, p, 3))
        x = torch.einsum("nhwpqc->nchpwq", x)
        imgs = x.reshape(shape=(x.shape[0], 3, h * p, h * p))

        return imgs

    def forward_encoder(self, x: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        # Tokens selected for masking at mask unit level
        num_windows = math.prod(self.encoder.mask_spatial_shape)

        (mask, _, _) = uniform_mask(x.size(0), num_windows, self.mask_ratio, device=x.device)

        # Get multi-scale representations from encoder
        (intermediates, mask) = self.encoder.masked_encoding(x, mask)

        # Resolution unchanged after q_pool stages, so skip those features
        intermediates = intermediates[: self.encoder.q_pool] + intermediates[-1:]

        # Multi-scale fusion
        x = 0.0
        for head, inter_x in zip(self.multi_scale_fusion_heads, intermediates):
            x += apply_fusion_head(head, inter_x)

        x = self.encoder_norm(x)

        return (x, mask)

    def forward_decoder(self, x: torch.Tensor, mask: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        x = self.decoder_embed(x)

        x_dec = torch.zeros(*mask.shape, *x.shape[2:], device=x.device, dtype=x.dtype)
        mask_tokens = self.mask_token.view((1,) * (len(mask.shape) + len(x.shape[2:-1])) + (-1,))
        mask = mask.reshape(mask.shape + (1,) * len(x.shape[2:]))
        mask = mask.expand((-1,) * 2 + x.shape[2:]).bool()
        x_dec[mask] = x.flatten()
        x_dec = ~mask * mask_tokens + mask * x_dec

        # Get back spatial order
        x = undo_windowing(
            x_dec,
            self.tokens_spatial_shape_final,  # type:ignore[arg-type]
            self.mask_unit_spatial_shape_final,
        )
        mask = undo_windowing(
            mask[..., 0:1],
            self.tokens_spatial_shape_final,  # type:ignore[arg-type]
            self.mask_unit_spatial_shape_final,
        )

        # Flatten
        x = x.reshape(x.shape[0], -1, x.shape[-1])
        mask = mask.view(x.shape[0], -1)

        # Add pos embed
        x = x + self.decoder_pos_embed

        # Apply decoder blocks
        x = self.decoder_blocks(x)
        x = self.decoder_norm(x)

        # Predictor projection
        x = self.decoder_pred(x)

        return (x, mask)

    def forward_loss(self, x: torch.Tensor, pred: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        label = self.get_pixel_label_2d(x, mask, norm=True)

        pred = pred[mask]
        loss = (pred - label) ** 2

        return loss.mean()

    def forward(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        (latent, mask) = self.forward_encoder(x)
        (pred, pred_mask) = self.forward_decoder(latent, mask)
        loss = self.forward_loss(x, pred, ~pred_mask)

        return {"loss": loss, "pred": pred, "mask": mask}


registry.register_weights(
    "mae_hiera_hiera_abswin_base",
    {
        "url": "https://huggingface.co/birder-project/mae_hiera_hiera_abswin_base/resolve/main",
        "description": "Masked auto-encoder Hiera with a Hiera abswin base image encoder, trained on 12M images",
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 328.1,
                "sha256": "bdf8a8495c99f4cb0f345cfcc5d616a427b59d46be0c5a7936766b60cd1efa60",
            }
        },
        "net": {"network": "mae_hiera"},
        "encoder": {"network": "hiera_abswin_base"},
    },
)



"""
TinyViT, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/tiny_vit.py

Paper "TinyViT: Fast Pretraining Distillation for Small Vision Transformers", https://arxiv.org/abs/2207.10666

Changes from original:
* Window sizes based on image size
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import interpolate_attention_bias


class PatchEmbed(nn.Module):
    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.conv1 = Conv2dNormActivation(
            in_channels,
            out_channels // 2,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv2 = Conv2dNormActivation(
            out_channels // 2, out_channels, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), activation_layer=None
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        return x


class MBConv(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, expand_ratio: float, drop_path: float) -> None:
        super().__init__()
        mid_channels = int(in_channels * expand_ratio)
        self.conv1 = Conv2dNormActivation(
            in_channels,
            mid_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv2 = Conv2dNormActivation(
            mid_channels,
            mid_channels,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            groups=mid_channels,
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv3 = Conv2dNormActivation(
            mid_channels,
            out_channels,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        shortcut = x
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.drop_path(x)
        x += shortcut

        return x


class ConvLayer(nn.Module):
    def __init__(self, dim: int, depth: int, drop_path: list[float], conv_expand_ratio: float) -> None:
        super().__init__()
        layers = []
        for i in range(depth):
            layers.append(MBConv(dim, dim, conv_expand_ratio, drop_path[i]))

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.blocks(x)


class PatchMerging(nn.Module):
    def __init__(self, dim: int, out_dim: int) -> None:
        super().__init__()
        self.conv1 = Conv2dNormActivation(
            dim,
            out_dim,
            kernel_size=(1, 1),
            stride=(1, 1),
            padding=(0, 0),
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv2 = Conv2dNormActivation(
            out_dim,
            out_dim,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            groups=out_dim,
            activation_layer=nn.GELU,
            inplace=None,
        )
        self.conv3 = Conv2dNormActivation(
            out_dim, out_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), activation_layer=None
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)

        return x


class NormMLP(nn.Module):
    def __init__(self, in_features: int, hidden_features: int, drop: float) -> None:
        super().__init__()
        self.norm = nn.LayerNorm(in_features)
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = nn.GELU()
        self.drop1 = nn.Dropout(drop)
        self.fc2 = nn.Linear(hidden_features, in_features)
        self.drop2 = nn.Dropout(drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x)
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop1(x)
        x = self.fc2(x)
        x = self.drop2(x)

        return x


class Attention(nn.Module):
    def __init__(self, dim: int, key_dim: int, num_heads: int, attn_ratio: float, resolution: tuple[int, int]) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = key_dim**-0.5
        self.key_dim = key_dim
        self.val_dim = int(attn_ratio * key_dim)
        self.out_dim = self.val_dim * num_heads

        self.norm = nn.LayerNorm(dim)
        self.qkv = nn.Linear(dim, num_heads * (self.val_dim + 2 * key_dim))
        self.proj = nn.Linear(self.out_dim, dim)

        self.define_bias_idxs(resolution)
        self.attention_biases = nn.Parameter(torch.zeros(self.num_heads, resolution[0] * resolution[1]))

    def define_bias_idxs(self, resolution: tuple[int, int]) -> None:
        k_pos = torch.stack(
            torch.meshgrid(torch.arange(resolution[0]), torch.arange(resolution[1]), indexing="ij")
        ).flatten(1)
        q_pos = torch.stack(
            torch.meshgrid(torch.arange(0, resolution[0]), torch.arange(0, resolution[1]), indexing="ij")
        ).flatten(1)
        rel_pos = (q_pos[..., :, None] - k_pos[..., None, :]).abs()
        rel_pos = (rel_pos[0] * resolution[1]) + rel_pos[1]
        self.attention_bias_idxs = nn.Buffer(torch.LongTensor(rel_pos), persistent=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        attn_bias = self.attention_biases[:, self.attention_bias_idxs]
        (B, N, _) = x.shape

        # Normalization
        x = self.norm(x)
        qkv = self.qkv(x)
        (q, k, v) = qkv.view(B, N, self.num_heads, -1).split([self.key_dim, self.key_dim, self.val_dim], dim=3)

        q = q.permute(0, 2, 1, 3)
        k = k.permute(0, 2, 1, 3)
        v = v.permute(0, 2, 1, 3)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=attn_bias, scale=self.scale
        )
        x = x.transpose(1, 2).reshape(B, N, self.out_dim)
        x = self.proj(x)

        return x


class TinyVitBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        window_size: tuple[int, int],
        mlp_ratio: float,
        drop: float,
        drop_path: list[float],
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim must be divisible by num_heads"

        self.dim = dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.mlp_ratio = mlp_ratio
        head_dim = dim // num_heads

        window_resolution = window_size
        self.attn = Attention(dim, head_dim, num_heads, attn_ratio=1, resolution=window_resolution)
        self.drop_path1 = StochasticDepth(drop_path, mode="row")

        self.mlp = NormMLP(dim, hidden_features=int(dim * mlp_ratio), drop=drop)
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

        self.local_conv = Conv2dNormActivation(
            dim, dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), groups=dim, activation_layer=None
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, H, W, C) = x.shape
        L = H * W

        shortcut = x
        if H == self.window_size[0] and W == self.window_size[1]:
            x = x.reshape(B, L, C)
            x = self.attn(x)
            x = x.view(B, H, W, C)
        else:
            pad_b = (self.window_size[0] - H % self.window_size[0]) % self.window_size[0]
            pad_r = (self.window_size[1] - W % self.window_size[1]) % self.window_size[1]
            padding = pad_b > 0 or pad_r > 0
            if padding:
                x = F.pad(x, (0, 0, 0, pad_r, 0, pad_b))

            # Window partition
            pH = H + pad_b  # pylint: disable=invalid-name
            pW = W + pad_r  # pylint: disable=invalid-name
            nH = pH // self.window_size[0]  # pylint: disable=invalid-name
            nW = pW // self.window_size[1]  # pylint: disable=invalid-name
            x = (
                x.view(B, nH, self.window_size[0], nW, self.window_size[1], C)
                .transpose(2, 3)
                .reshape(B * nH * nW, self.window_size[0] * self.window_size[1], C)
            )

            x = self.attn(x)

            # Window reverse
            x = x.view(B, nH, nW, self.window_size[0], self.window_size[1], C).transpose(2, 3).reshape(B, pH, pW, C)

            if padding:
                x = x[:, :H, :W].contiguous()

        x = shortcut + self.drop_path1(x)

        x = x.permute(0, 3, 1, 2)
        x = self.local_conv(x)
        x = x.reshape(B, C, L).transpose(1, 2)
        x = x + self.drop_path2(self.mlp(x))

        return x.view(B, H, W, C)


class TinyVitStage(nn.Module):
    def __init__(
        self,
        dim: int,
        out_dim: int,
        depth: int,
        num_heads: int,
        window_size: tuple[int, int],
        mlp_ratio: float,
        drop: float,
        drop_path: list[list[float]],
        downsample: bool,
    ) -> None:
        super().__init__()
        self.depth = depth
        self.out_dim = out_dim

        if downsample is True:
            self.downsample = PatchMerging(dim=dim, out_dim=out_dim)
        else:
            self.downsample = nn.Identity()
            assert dim == out_dim

        layers = []
        for i in range(depth):
            layers.append(
                TinyVitBlock(
                    dim=out_dim,
                    num_heads=num_heads,
                    window_size=window_size,
                    mlp_ratio=mlp_ratio,
                    drop=drop,
                    drop_path=drop_path[i],
                )
            )

        self.blocks = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.downsample(x)
        x = x.permute(0, 2, 3, 1)  # BCHW -> BHWC
        x = self.blocks(x)
        x = x.permute(0, 3, 1, 2)  # BHWC -> BCHW

        return x


# pylint: disable=invalid-name
class Tiny_ViT(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        self.window_scale_factors = [1, 1, 2, 1]
        embed_dims: list[int] = self.config["embed_dims"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]
        window_sizes = [
            (int(self.size[0] / (2**5) * scale), int(self.size[1] / (2**5) * scale))
            for scale in self.window_scale_factors
        ]

        self.stem = PatchEmbed(in_channels=self.input_channels, out_channels=embed_dims[0])

        num_stages = len(depths)
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        prev_dim = embed_dims[0]
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for stage_idx in range(num_stages):
            if stage_idx == 0:
                stages[f"stage{stage_idx+1}"] = ConvLayer(
                    prev_dim,
                    depth=depths[stage_idx],
                    drop_path=dpr[: depths[stage_idx]],
                    conv_expand_ratio=4.0,
                )
            else:
                out_dim = embed_dims[stage_idx]
                stages[f"stage{stage_idx+1}"] = TinyVitStage(
                    dim=embed_dims[stage_idx - 1],
                    out_dim=out_dim,
                    depth=depths[stage_idx],
                    num_heads=num_heads[stage_idx],
                    window_size=window_sizes[stage_idx],
                    mlp_ratio=4.0,
                    drop=0.0,
                    drop_path=dpr[sum(depths[:stage_idx]) : sum(depths[: stage_idx + 1])],
                    downsample=True,
                )
                prev_dim = out_dim

            return_channels.append(prev_dim)

        num_features = embed_dims[-1]
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(num_features),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        old_window_sizes = [
            (int(old_size[0] / (2**5) * scale), int(old_size[1] / (2**5) * scale))
            for scale in self.window_scale_factors
        ]
        window_sizes = [
            (int(new_size[0] / (2**5) * scale), int(new_size[1] / (2**5) * scale))
            for scale in self.window_scale_factors
        ]
        idx = 0
        for stage in self.body:
            if isinstance(stage, TinyVitStage):
                for m in stage.modules():
                    if isinstance(m, TinyVitBlock):
                        m.window_size = window_sizes[idx]

                        # This will update the index buffer
                        m.attn.define_bias_idxs(window_sizes[idx])

                        # Interpolate the actual table
                        m.attn.attention_biases = nn.Parameter(
                            interpolate_attention_bias(
                                m.attn.attention_biases, old_window_sizes[idx], window_sizes[idx], mode="bilinear"
                            )
                        )

            idx += 1


registry.register_model_config(
    "tiny_vit_5m",
    Tiny_ViT,
    config={
        "embed_dims": [64, 128, 160, 320],
        "depths": [2, 2, 6, 2],
        "num_heads": [2, 4, 5, 10],
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "tiny_vit_11m",
    Tiny_ViT,
    config={
        "embed_dims": [64, 128, 256, 448],
        "depths": [2, 2, 6, 2],
        "num_heads": [2, 4, 8, 14],
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "tiny_vit_21m",
    Tiny_ViT,
    config={
        "embed_dims": [96, 192, 384, 576],
        "depths": [2, 2, 6, 2],
        "num_heads": [3, 6, 12, 18],
        "drop_path_rate": 0.2,
    },
)

registry.register_weights(
    "tiny_vit_5m_il-common",
    {
        "description": "TinyViT 5M model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 20.0,
                "sha256": "57f84dc3144fc4e3ca39328d3a1446ca9e26ddb54e4c4d84301b7638bee2ec21",
            },
        },
        "net": {"network": "tiny_vit_5m", "tag": "il-common"},
    },
)



import argparse
import logging
import os
from typing import Any

import torch
from torch.utils.data import DataLoader
from torchvision.datasets import CocoDetection
from torchvision.datasets import wrap_dataset_for_transforms_v2
from torchvision.io import decode_image
from torchvision.transforms import v2
from tqdm import tqdm

from birder.common import cli
from birder.conf import settings
from birder.data.collators.detection import collate_fn

logger = logging.getLogger(__name__)


def verify_coco(args: argparse.Namespace) -> None:
    batch_size = 32

    transform = v2.Compose(
        [
            v2.Resize((256, 256), interpolation=v2.InterpolationMode.BILINEAR),
            v2.PILToTensor(),
            v2.ToDtype(torch.float32, scale=True),
        ]
    )

    dataset = CocoDetection(args.data_path, args.coco_json_path, transforms=transform)
    dataset = wrap_dataset_for_transforms_v2(dataset)
    total = len(dataset)

    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=8,
        collate_fn=collate_fn,
        drop_last=False,
    )

    with tqdm(total=total, initial=0, unit="images", unit_scale=True, leave=False) as progress:
        if args.fast is True:
            idx = 0
            try:
                for idx, (_, _) in enumerate(data_loader):
                    progress.update(batch_size)

            except (OSError, RuntimeError) as e:
                logger.warning(
                    f"File at batch no. {idx} (batch size = {batch_size}, "
                    f"{(idx-1) * batch_size}) failed to load {e}"
                )

        else:
            for img_id in dataset.coco.imgs:
                try:
                    img_path = dataset.coco.loadImgs(img_id)[0]["file_name"]
                    img_path = os.path.join(dataset.root, img_path)
                    img = decode_image(img_path)
                    img = transform(img)
                    if img.size(0) != 3:
                        logger.warning(f"File {img_path} failed to load {img.size()}")

                except (OSError, RuntimeError) as e:
                    logger.warning(f"File {img_path} failed to load {e}")

                progress.update(1)

    logger.info("Finished")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "verify-coco",
        allow_abbrev=False,
        help="loads every image in the dataset and raises exception on corrupted or missing files",
        description="loads every image in the dataset and raises exception on corrupted or missing files",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools verify-coco\n"
            "python -m birder.tools verify-coco --coco-json-path "
            "data/detection_data/training_annotations_coco.json --data-path data/detection_data\n"
            "python -m birder.tools verify-coco --fast --coco-json-path "
            "~/Datasets/Objects365-2020/train/zhiyuan_objv2_train.json --data-path ~/Datasets/Objects365-2020/train\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--fast", default=False, action="store_true", help="use parallel dataloader")
    subparser.add_argument(
        "--data-path",
        type=str,
        default=str(settings.DETECTION_DATA_PATH),
        help="image directory path",
    )
    subparser.add_argument(
        "--coco-json-path",
        type=str,
        default=f"{settings.TRAINING_DETECTION_ANNOTATIONS_PATH}_coco.json",
        help="training COCO json path",
    )
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    verify_coco(args)



"""
CoaT, adapted from
https://github.com/huggingface/pytorch-image-models/blob/main/timm/models/coat.py

Paper "Co-Scale Conv-Attentional Image Transformers", https://arxiv.org/abs/2104.06399
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


def insert_cls(x: torch.Tensor, cls_token: torch.Tensor) -> torch.Tensor:
    cls_tokens = cls_token.expand(x.shape[0], -1, -1)
    x = torch.concat((cls_tokens, x), dim=1)

    return x


def remove_cls(x: torch.Tensor) -> torch.Tensor:
    return x[:, 1:, :]


class ConvRelPosEnc(nn.Module):
    def __init__(self, head_channels: int, window: dict[int, int]) -> None:
        super().__init__()

        self.conv_list = nn.ModuleList()
        head_splits = []
        for cur_window, cur_head_split in window.items():
            dilation = 1
            # Determine padding size.
            # Ref: https://discuss.pytorch.org/t/how-to-keep-the-shape-of-input-and-output-same-when-dilation-conv/14338
            padding_size = (cur_window + (cur_window - 1) * (dilation - 1)) // 2
            cur_conv = nn.Conv2d(
                cur_head_split * head_channels,
                cur_head_split * head_channels,
                kernel_size=(cur_window, cur_window),
                stride=(1, 1),
                padding=(padding_size, padding_size),
                dilation=(dilation, dilation),
                groups=cur_head_split * head_channels,
            )
            self.conv_list.append(cur_conv)
            head_splits.append(cur_head_split)

        self.channel_splits = [x * head_channels for x in head_splits]

    def forward(self, q: torch.Tensor, v: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, num_heads, N, C) = q.size()
        (H, W) = size
        torch._assert(N == 1 + H * W, "size mismatch")  # pylint: disable=protected-access

        # Convolutional relative position encoding.
        q_img = q[:, :, 1:, :]  # [B, h, H*W, Ch]
        v_img = v[:, :, 1:, :]  # [B, h, H*W, Ch]

        v_img = v_img.transpose(-1, -2).reshape(B, num_heads * C, H, W)
        v_img_list = torch.split(v_img, self.channel_splits, dim=1)  # Split according to channels
        conv_v_img_list = []
        for i, conv in enumerate(self.conv_list):
            conv_v_img_list.append(conv(v_img_list[i]))

        conv_v_img = torch.concat(conv_v_img_list, dim=1)
        conv_v_img = conv_v_img.reshape(B, num_heads, C, H * W).transpose(-1, -2)

        ev_hat = q_img * conv_v_img
        ev_hat = F.pad(ev_hat, (0, 0, 1, 0, 0, 0))

        return ev_hat


class FactorAttnConvRelPosEnc(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        qkv_bias: bool,
        proj_drop: float,
        shared_crpe: nn.Module,
    ) -> None:
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        # Shared convolutional relative position encoding
        self.crpe = shared_crpe

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, N, C) = x.size()

        # Generate Q, K, V
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        (q, k, v) = qkv.unbind(0)  # [B, h, N, Ch]

        # Factorized attention
        k_softmax = k.softmax(dim=2)
        factor_att = k_softmax.transpose(-1, -2) @ v
        factor_att = q @ factor_att

        # Convolutional relative position encoding
        crpe = self.crpe(q, v, size=size)  # [B, h, N, Ch]

        # Merge and reshape
        x = self.scale * factor_att + crpe
        x = x.transpose(1, 2).reshape(B, N, C)  # [B, h, N, Ch] -> [B, N, h, Ch] -> [B, N, C]

        # Output projection
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class ConvPosEnc(nn.Module):
    def __init__(self, dim: int, kernel_size: int):
        super().__init__()
        self.proj = nn.Conv2d(
            dim, dim, kernel_size=kernel_size, stride=(1, 1), padding=(kernel_size // 2, kernel_size // 2), groups=dim
        )

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        (B, N, C) = x.size()
        (H, W) = size
        torch._assert(N == 1 + H * W, "size mismatch")  # pylint: disable=protected-access

        # Extract CLS token and image tokens
        cls_token = x[:, :1]  # [B, 1, C]
        img_tokens = x[:, 1:]  # [B, H*W, C]

        # Depthwise convolution
        feat = img_tokens.transpose(1, 2).view(B, C, H, W)
        x = self.proj(feat) + feat
        x = x.flatten(2).transpose(1, 2)

        # Combine with CLS token
        x = torch.concat((cls_token, x), dim=1)

        return x


class SerialBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float,
        qkv_bias: bool,
        proj_drop: float,
        drop_path: float,
        shared_cpe: nn.Module,
        shared_crpe: nn.Module,
    ) -> None:
        super().__init__()

        # Conv-attention
        self.cpe = shared_cpe
        self.norm1 = nn.LayerNorm(dim)
        self.factor_attn_crpe = FactorAttnConvRelPosEnc(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            proj_drop=proj_drop,
            shared_crpe=shared_crpe,
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

        # MLP
        self.norm2 = nn.LayerNorm(dim)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, dropout=proj_drop)

    def forward(self, x: torch.Tensor, size: tuple[int, int]) -> torch.Tensor:
        # Conv-attention
        x = self.cpe(x, size)
        cur = self.norm1(x)
        cur = self.factor_attn_crpe(cur, size)
        x = x + self.drop_path(cur)

        # MLP
        cur = self.norm2(x)
        cur = self.mlp(cur)
        x = x + self.drop_path(cur)

        return x


class ParallelBlock(nn.Module):
    def __init__(
        self,
        dims: list[int],
        num_heads: int,
        mlp_ratios: list[float],
        qkv_bias: bool,
        proj_drop: float,
        drop_path: float,
        shared_crpes: list[nn.Module],
    ) -> None:
        super().__init__()

        # Conv-attention
        self.norm12 = nn.LayerNorm(dims[1])
        self.norm13 = nn.LayerNorm(dims[2])
        self.norm14 = nn.LayerNorm(dims[3])
        self.factor_attn_crpe2 = FactorAttnConvRelPosEnc(
            dims[1], num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop, shared_crpe=shared_crpes[1]
        )
        self.factor_attn_crpe3 = FactorAttnConvRelPosEnc(
            dims[2], num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop, shared_crpe=shared_crpes[2]
        )
        self.factor_attn_crpe4 = FactorAttnConvRelPosEnc(
            dims[3], num_heads=num_heads, qkv_bias=qkv_bias, proj_drop=proj_drop, shared_crpe=shared_crpes[3]
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

        # MLP
        self.norm22 = nn.LayerNorm(dims[1])
        self.norm23 = nn.LayerNorm(dims[2])
        self.norm24 = nn.LayerNorm(dims[3])

        # In the parallel block, we assume dimensions are the same and share the linear transformation
        assert dims[1] == dims[2] == dims[3]
        assert mlp_ratios[1] == mlp_ratios[2] == mlp_ratios[3]
        self.mlp = MLP(dims[1], [int(dims[1] * mlp_ratios[1]), dims[1]], activation_layer=nn.GELU, dropout=proj_drop)

    def upsample(self, x: torch.Tensor, factor: float, size: tuple[int, int]) -> torch.Tensor:
        return self.interpolate(x, scale_factor=factor, size=size)

    def downsample(self, x: torch.Tensor, factor: float, size: tuple[int, int]) -> torch.Tensor:
        return self.interpolate(x, scale_factor=1.0 / factor, size=size)

    def interpolate(self, x: torch.Tensor, scale_factor: float, size: tuple[int, int]) -> torch.Tensor:
        (B, N, C) = x.size()
        (H, W) = size
        torch._assert(N == 1 + H * W, "size mismatch")  # pylint: disable=protected-access

        cls_token = x[:, :1, :]
        img_tokens = x[:, 1:, :]

        img_tokens = img_tokens.transpose(1, 2).reshape(B, C, H, W)
        img_tokens = F.interpolate(
            img_tokens,
            scale_factor=scale_factor,
            recompute_scale_factor=False,
            mode="bilinear",
            align_corners=False,
        )
        img_tokens = img_tokens.reshape(B, C, -1).transpose(1, 2)

        out = torch.concat((cls_token, img_tokens), dim=1)

        return out

    def forward(
        self, x1: torch.Tensor, x2: torch.Tensor, x3: torch.Tensor, x4: torch.Tensor, sizes: list[tuple[int, int]]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        (_, s2, s3, s4) = sizes
        cur2 = self.norm12(x2)
        cur3 = self.norm13(x3)
        cur4 = self.norm14(x4)
        cur2 = self.factor_attn_crpe2(cur2, size=s2)
        cur3 = self.factor_attn_crpe3(cur3, size=s3)
        cur4 = self.factor_attn_crpe4(cur4, size=s4)
        upsample3_2 = self.upsample(cur3, factor=2.0, size=s3)
        upsample4_3 = self.upsample(cur4, factor=2.0, size=s4)
        upsample4_2 = self.upsample(cur4, factor=4.0, size=s4)
        downsample2_3 = self.downsample(cur2, factor=2.0, size=s2)
        downsample3_4 = self.downsample(cur3, factor=2.0, size=s3)
        downsample2_4 = self.downsample(cur2, factor=4.0, size=s2)
        cur2 = cur2 + upsample3_2 + upsample4_2
        cur3 = cur3 + upsample4_3 + downsample2_3
        cur4 = cur4 + downsample3_4 + downsample2_4
        x2 = x2 + self.drop_path(cur2)
        x3 = x3 + self.drop_path(cur3)
        x4 = x4 + self.drop_path(cur4)

        # MLP
        cur2 = self.norm22(x2)
        cur3 = self.norm23(x3)
        cur4 = self.norm24(x4)
        cur2 = self.mlp(cur2)
        cur3 = self.mlp(cur3)
        cur4 = self.mlp(cur4)
        x2 = x2 + self.drop_path(cur2)
        x3 = x3 + self.drop_path(cur3)
        x4 = x4 + self.drop_path(cur4)

        return (x1, x2, x3, x4)


class PatchEmbed(nn.Module):
    def __init__(self, patch_size: tuple[int, int], in_channels: int, embed_dim: int) -> None:
        super().__init__()
        self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, padding=(0, 0))
        self.norm = nn.LayerNorm(embed_dim)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor, tuple[int, int]]:
        x = self.proj(x)
        (H, W) = x.shape[2:4]

        x = x.flatten(2).transpose(1, 2)
        x = self.norm(x)

        return (x, (H, W))


# pylint: disable=too-many-instance-attributes
class CoaT(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = (4, 4)
        crpe_window = {3: 2, 5: 3, 7: 3}
        num_heads = 8
        qkv_bias = True
        proj_drop_rate = 0.0
        embed_dims: list[int] = self.config["embed_dims"]
        mlp_ratios: list[float] = self.config["mlp_ratios"]
        serial_depths: list[int] = self.config["serial_depths"]
        parallel_depth: int = self.config["parallel_depth"]
        drop_path_rate: float = self.config["drop_path_rate"]  # Not using the "dpr rule"

        # Patch embeddings
        self.patch_embed1 = PatchEmbed(patch_size=patch_size, in_channels=self.input_channels, embed_dim=embed_dims[0])
        self.patch_embed2 = PatchEmbed(patch_size=(2, 2), in_channels=embed_dims[0], embed_dim=embed_dims[1])
        self.patch_embed3 = PatchEmbed(patch_size=(2, 2), in_channels=embed_dims[1], embed_dim=embed_dims[2])
        self.patch_embed4 = PatchEmbed(patch_size=(2, 2), in_channels=embed_dims[2], embed_dim=embed_dims[3])

        # Class tokens
        self.cls_token1 = nn.Parameter(torch.zeros(1, 1, embed_dims[0]))
        self.cls_token2 = nn.Parameter(torch.zeros(1, 1, embed_dims[1]))
        self.cls_token3 = nn.Parameter(torch.zeros(1, 1, embed_dims[2]))
        self.cls_token4 = nn.Parameter(torch.zeros(1, 1, embed_dims[3]))

        # Convolutional position encodings
        self.cpe1 = ConvPosEnc(dim=embed_dims[0], kernel_size=3)
        self.cpe2 = ConvPosEnc(dim=embed_dims[1], kernel_size=3)
        self.cpe3 = ConvPosEnc(dim=embed_dims[2], kernel_size=3)
        self.cpe4 = ConvPosEnc(dim=embed_dims[3], kernel_size=3)

        # Convolutional relative position encodings
        self.crpe1 = ConvRelPosEnc(head_channels=embed_dims[0] // num_heads, window=crpe_window)
        self.crpe2 = ConvRelPosEnc(head_channels=embed_dims[1] // num_heads, window=crpe_window)
        self.crpe3 = ConvRelPosEnc(head_channels=embed_dims[2] // num_heads, window=crpe_window)
        self.crpe4 = ConvRelPosEnc(head_channels=embed_dims[3] // num_heads, window=crpe_window)

        # Serial blocks
        self.serial_blocks1 = nn.ModuleList()
        self.serial_blocks2 = nn.ModuleList()
        self.serial_blocks3 = nn.ModuleList()
        self.serial_blocks4 = nn.ModuleList()
        for _ in range(serial_depths[0]):
            self.serial_blocks1.append(
                SerialBlock(
                    dim=embed_dims[0],
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratios[0],
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop_rate,
                    drop_path=drop_path_rate,
                    shared_cpe=self.cpe1,
                    shared_crpe=self.crpe1,
                )
            )

        for _ in range(serial_depths[1]):
            self.serial_blocks2.append(
                SerialBlock(
                    dim=embed_dims[1],
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratios[1],
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop_rate,
                    drop_path=drop_path_rate,
                    shared_cpe=self.cpe2,
                    shared_crpe=self.crpe2,
                )
            )

        for _ in range(serial_depths[2]):
            self.serial_blocks3.append(
                SerialBlock(
                    dim=embed_dims[2],
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratios[2],
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop_rate,
                    drop_path=drop_path_rate,
                    shared_cpe=self.cpe3,
                    shared_crpe=self.crpe3,
                )
            )

        for _ in range(serial_depths[3]):
            self.serial_blocks4.append(
                SerialBlock(
                    dim=embed_dims[3],
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratios[3],
                    qkv_bias=qkv_bias,
                    proj_drop=proj_drop_rate,
                    drop_path=drop_path_rate,
                    shared_cpe=self.cpe4,
                    shared_crpe=self.crpe4,
                )
            )

        # Parallel blocks
        if parallel_depth > 0:
            self.parallel_blocks = nn.ModuleList()
            for _ in range(parallel_depth):
                self.parallel_blocks.append(
                    ParallelBlock(
                        dims=embed_dims,
                        num_heads=num_heads,
                        mlp_ratios=mlp_ratios,
                        qkv_bias=qkv_bias,
                        proj_drop=proj_drop_rate,
                        drop_path=drop_path_rate,
                        shared_crpes=[self.crpe1, self.crpe2, self.crpe3, self.crpe4],
                    )
                )
        else:
            self.parallel_blocks = None

        # Norms
        if self.parallel_blocks is not None:
            self.norm2 = nn.LayerNorm(embed_dims[1])
            self.norm3 = nn.LayerNorm(embed_dims[2])
        else:
            self.norm2 = None
            self.norm3 = None

        self.norm4 = nn.LayerNorm(embed_dims[3])

        # Head
        if parallel_depth > 0:
            # CoaT series: aggregate features of last three scales for classification.
            assert embed_dims[1] == embed_dims[2] == embed_dims[3]
            self.aggregate = nn.Conv1d(in_channels=3, out_channels=1, kernel_size=1, stride=1, padding=0)
        else:
            # CoaT-Lite series: use feature of last scale for classification.
            self.aggregate = None

        self.return_channels = embed_dims[1:]
        self.return_stages = self.return_stages[1:]
        self.embedding_size = embed_dims[-1]
        self.classifier = self.create_classifier()

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        nn.init.trunc_normal_(self.cls_token1, std=0.02)
        nn.init.trunc_normal_(self.cls_token2, std=0.02)
        nn.init.trunc_normal_(self.cls_token3, std=0.02)
        nn.init.trunc_normal_(self.cls_token4, std=0.02)

    def transform_to_backbone(self) -> None:
        if self.norm2 is not None:
            self.norm2 = nn.Identity()
        if self.norm3 is not None:
            self.norm3 = nn.Identity()
        if self.aggregate is not None:
            self.aggregate = nn.Identity()

        self.norm4 = nn.Identity()
        self.classifier = nn.Identity()

    def _features(self, x: torch.Tensor) -> dict[str, tuple[torch.Tensor, torch.Tensor]]:
        B = x.shape[0]

        # Serial blocks 1
        (x1, (h1, w1)) = self.patch_embed1(x)
        x1 = insert_cls(x1, self.cls_token1)
        for blk in self.serial_blocks1:
            x1 = blk(x1, size=(h1, w1))

        x1_no_cls = remove_cls(x1).reshape(B, h1, w1, -1).permute(0, 3, 1, 2).contiguous()

        # Serial blocks 2
        (x2, (h2, w2)) = self.patch_embed2(x1_no_cls)
        x2 = insert_cls(x2, self.cls_token2)
        for blk in self.serial_blocks2:
            x2 = blk(x2, size=(h2, w2))

        x2_no_cls = remove_cls(x2).reshape(B, h2, w2, -1).permute(0, 3, 1, 2).contiguous()

        # Serial blocks 3
        (x3, (h3, w3)) = self.patch_embed3(x2_no_cls)
        x3 = insert_cls(x3, self.cls_token3)
        for blk in self.serial_blocks3:
            x3 = blk(x3, size=(h3, w3))

        x3_no_cls = remove_cls(x3).reshape(B, h3, w3, -1).permute(0, 3, 1, 2).contiguous()

        # Serial blocks 4
        (x4, (h4, w4)) = self.patch_embed4(x3_no_cls)
        x4 = insert_cls(x4, self.cls_token4)
        for blk in self.serial_blocks4:
            x4 = blk(x4, size=(h4, w4))

        x4_no_cls = remove_cls(x4).reshape(B, h4, w4, -1).permute(0, 3, 1, 2).contiguous()

        if self.parallel_blocks is not None:
            # Parallel blocks
            for blk in self.parallel_blocks:
                x2 = self.cpe2(x2, (h2, w2))
                x3 = self.cpe3(x3, (h3, w3))
                x4 = self.cpe4(x4, (h4, w4))
                (x1, x2, x3, x4) = blk(x1, x2, x3, x4, sizes=[(h1, w1), (h2, w2), (h3, w3), (h4, w4)])

            x1_no_cls = remove_cls(x1).reshape(B, h1, w1, -1).permute(0, 3, 1, 2).contiguous()
            x2_no_cls = remove_cls(x2).reshape(B, h2, w2, -1).permute(0, 3, 1, 2).contiguous()
            x3_no_cls = remove_cls(x3).reshape(B, h3, w3, -1).permute(0, 3, 1, 2).contiguous()
            x4_no_cls = remove_cls(x4).reshape(B, h4, w4, -1).permute(0, 3, 1, 2).contiguous()

        return {
            "stage1": (x1, x1_no_cls),
            "stage2": (x2, x2_no_cls),
            "stage3": (x3, x3_no_cls),
            "stage4": (x4, x4_no_cls),
        }

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        features = self._features(x)
        out = {}
        for name, feat in features.items():
            if name in self.return_stages:
                out[name] = feat[1]

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        if up_to_stage >= 1:
            for param in self.patch_embed1.parameters():
                param.requires_grad = False
            for param in self.serial_blocks1.parameters():
                param.requires_grad = False
        if up_to_stage >= 2:
            for param in self.patch_embed2.parameters():
                param.requires_grad = False
            for param in self.serial_blocks2.parameters():
                param.requires_grad = False
        if up_to_stage >= 3:
            for param in self.patch_embed3.parameters():
                param.requires_grad = False
            for param in self.serial_blocks3.parameters():
                param.requires_grad = False
        if up_to_stage >= 4:
            for param in self.patch_embed4.parameters():
                param.requires_grad = False
            for param in self.serial_blocks4.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        features = self._features(x)
        return features["stage4"][0]

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        features = self._features(x)
        if self.parallel_blocks is None:
            x4 = features["stage4"][0]
            x4 = self.norm4(x4)
            x4_cls = x4[:, 0]
            return x4_cls

        x2 = self.norm2(features["stage2"][0])
        x3 = self.norm3(features["stage3"][0])
        x4 = self.norm4(features["stage4"][0])

        x2_cls = x2[:, :1]
        x3_cls = x3[:, :1]
        x4_cls = x4[:, :1]

        merged_cls = torch.concat((x2_cls, x3_cls, x4_cls), dim=1)
        merged_cls = self.aggregate(merged_cls).squeeze(dim=1)

        return merged_cls


registry.register_model_config(
    "coat_tiny",
    CoaT,
    config={
        "embed_dims": [152, 152, 152, 152],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "serial_depths": [2, 2, 2, 2],
        "parallel_depth": 6,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "coat_mini",
    CoaT,
    config={
        "embed_dims": [152, 216, 216, 216],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "serial_depths": [2, 2, 2, 2],
        "parallel_depth": 6,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "coat_small",
    CoaT,
    config={
        "embed_dims": [152, 320, 320, 320],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "serial_depths": [2, 2, 2, 2],
        "parallel_depth": 6,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "coat_lite_tiny",
    CoaT,
    config={
        "embed_dims": [64, 128, 256, 320],
        "mlp_ratios": [8.0, 8.0, 4.0, 4.0],
        "serial_depths": [2, 2, 2, 2],
        "parallel_depth": 0,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "coat_lite_mini",
    CoaT,
    config={
        "embed_dims": [64, 128, 320, 512],
        "mlp_ratios": [8.0, 8.0, 4.0, 4.0],
        "serial_depths": [2, 2, 2, 2],
        "parallel_depth": 0,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "coat_lite_small",
    CoaT,
    config={
        "embed_dims": [64, 128, 320, 512],
        "mlp_ratios": [8.0, 8.0, 4.0, 4.0],
        "serial_depths": [3, 4, 6, 3],
        "parallel_depth": 0,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "coat_lite_medium",
    CoaT,
    config={
        "embed_dims": [128, 256, 320, 512],
        "mlp_ratios": [4.0, 4.0, 4.0, 4.0],
        "serial_depths": [3, 6, 10, 8],
        "parallel_depth": 0,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "coat_tiny_il-common",
    {
        "description": "CoaT tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 20.8,
                "sha256": "d6bef0c932233a48d297855e04ca421484672e4c684581c486d3cb7b2fcd0003",
            }
        },
        "net": {"network": "coat_tiny", "tag": "il-common"},
    },
)
registry.register_weights(
    "coat_lite_tiny_il-common",
    {
        "description": "CoaT lite tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 21.1,
                "sha256": "275edeeccaa74547056a9a3ee09dd36b17c7c97e453030b6f62b18dd02107bf1",
            }
        },
        "net": {"network": "coat_lite_tiny", "tag": "il-common"},
    },
)



import logging
import unittest

from birder.data.datasets import directory
from birder.data.datasets import webdataset

logging.disable(logging.CRITICAL)


class TestDatasets(unittest.TestCase):
    def test_directory(self) -> None:
        dataset = directory.ImageListDataset(
            [("file1.jpeg", 0), ("file2.jpeg", 1), ("file3.jpeg", 0), ("file4.jpeg", 0)],
            transforms=lambda x: x + ".data",
            loader=lambda x: x,
        )

        self.assertEqual(len(dataset), 4)
        (path, sample, label) = dataset[2]
        self.assertEqual(path, "file3.jpeg")
        self.assertEqual(sample, "file3.jpeg.data")
        self.assertEqual(label, 0)

        repr(dataset)

    def test_webdataset(self) -> None:
        (sample_name, data, label) = webdataset.decode_sample_name(("shard1", "sample6", b"data", 1))
        self.assertEqual(sample_name, "shard1/sample6")
        self.assertEqual(data, b"data")
        self.assertEqual(label, 1)



"""
Deformable DETR, adapted from
https://github.com/fundamentalvision/Deformable-DETR/blob/main/models/deformable_detr.py
and
https://github.com/huggingface/transformers/blob/main/src/transformers/models/deformable_detr/modeling_deformable_detr.py

Paper "Deformable DETR: Deformable Transformers for End-to-End Object Detection",
https://arxiv.org/abs/2010.04159

Changes from original:
* Removed two stage support
* Zero cost matrix elements on overflow (HungarianMatcher)
"""

# Reference license: Apache-2.0 (both)

import copy
import math
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import boxes as box_ops
from torchvision.ops import sigmoid_focal_loss

from birder.common import training_utils
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.detr import PositionEmbeddingSine
from birder.ops.msda import MultiScaleDeformableAttention as MSDA
from birder.ops.soft_nms import SoftNMS


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class HungarianMatcher(nn.Module):
    """
    This class computes an assignment between the targets and the predictions of the network
    """

    def __init__(self, cost_class: float, cost_bbox: float, cost_giou: float):
        super().__init__()
        assert cost_class != 0 or cost_bbox != 0 or cost_giou != 0
        self.cost_class = cost_class
        self.cost_bbox = cost_bbox
        self.cost_giou = cost_giou

    @torch.jit.unused  # type: ignore[misc]
    def forward(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, targets: list[dict[str, torch.Tensor]]
    ) -> list[torch.Tensor]:
        with torch.no_grad():
            (B, num_queries) = class_logits.shape[:2]

            # We flatten to compute the cost matrices in a batch
            out_prob = class_logits.flatten(0, 1).sigmoid()  # [batch_size * num_queries, num_classes]
            out_bbox = box_regression.flatten(0, 1)  # [batch_size * num_queries, 4]

            # Also concat the target labels and boxes
            tgt_ids = torch.concat([v["labels"] for v in targets], dim=0)
            tgt_bbox = torch.concat([v["boxes"] for v in targets], dim=0)

            # Compute the classification cost
            alpha = 0.25
            gamma = 2.0
            neg_cost_class = (1 - alpha) * (out_prob**gamma) * (-(1 - out_prob + 1e-8).log())
            pos_cost_class = alpha * ((1 - out_prob) ** gamma) * (-(out_prob + 1e-8).log())
            cost_class = pos_cost_class[:, tgt_ids] - neg_cost_class[:, tgt_ids]

            # Compute the L1 cost between boxes
            cost_bbox = torch.cdist(out_bbox, tgt_bbox, p=1.0)

            # Compute the giou cost between boxes
            cost_giou = -box_ops.generalized_box_iou(
                box_ops.box_convert(out_bbox, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(tgt_bbox, in_fmt="cxcywh", out_fmt="xyxy"),
            )

            # Final cost matrix
            C = self.cost_bbox * cost_bbox + self.cost_class * cost_class + self.cost_giou * cost_giou
            C = C.view(B, num_queries, -1).cpu()
            C[C.isnan() | C.isinf()] = 0.0

            sizes = [len(v["boxes"]) for v in targets]
            indices = [linear_sum_assignment(c[i]) for i, c in enumerate(C.split(sizes, -1))]

            return [(torch.as_tensor(i, dtype=torch.int64), torch.as_tensor(j, dtype=torch.int64)) for i, j in indices]


def inverse_sigmoid(x: torch.Tensor, eps: float = 1e-5) -> torch.Tensor:
    x = x.clamp(min=0, max=1)
    x1 = x.clamp(min=eps)
    x2 = (1 - x).clamp(min=eps)

    return torch.log(x1 / x2)


class MultiScaleDeformableAttention(nn.Module):
    def __init__(self, d_model: int, n_levels: int, n_heads: int, n_points: int) -> None:
        super().__init__()
        if d_model % n_heads != 0:
            raise ValueError(f"d_model must be divisible by n_heads, but got {d_model} and {n_heads}")

        # Ensure dim_per_head is power of 2
        dim_per_head = d_model // n_heads
        if ((dim_per_head & (dim_per_head - 1) == 0) and dim_per_head != 0) is False:
            raise ValueError(
                "Set d_model in MultiScaleDeformableAttention to make the dimension of each attention head a power of 2"
            )

        self.im2col_step = 64
        self.d_model = d_model
        self.n_levels = n_levels
        self.n_heads = n_heads
        self.n_points = n_points

        self.msda = MSDA()
        self.sampling_offsets = nn.Linear(d_model, n_heads * n_levels * n_points * 2)
        self.attention_weights = nn.Linear(d_model, n_heads * n_levels * n_points)
        self.value_proj = nn.Linear(d_model, d_model)
        self.output_proj = nn.Linear(d_model, d_model)

        self._reset_parameters()

    def _reset_parameters(self) -> None:
        nn.init.constant_(self.sampling_offsets.weight.data, 0.0)
        thetas = torch.arange(self.n_heads, dtype=torch.float32) * (2.0 * math.pi / self.n_heads)
        grid_init = torch.stack([thetas.cos(), thetas.sin()], -1)
        grid_init = (
            (grid_init / grid_init.abs().max(-1, keepdim=True)[0])
            .view(self.n_heads, 1, 1, 2)
            .repeat(1, self.n_levels, self.n_points, 1)
        )
        for i in range(self.n_points):
            grid_init[:, :, i, :] *= i + 1

        with torch.no_grad():
            self.sampling_offsets.bias = nn.Parameter(grid_init.view(-1))

        nn.init.constant_(self.attention_weights.weight.data, 0.0)
        nn.init.constant_(self.attention_weights.bias.data, 0.0)
        nn.init.xavier_uniform_(self.value_proj.weight.data)
        nn.init.constant_(self.value_proj.bias.data, 0.0)
        nn.init.xavier_uniform_(self.output_proj.weight.data)
        nn.init.constant_(self.output_proj.bias.data, 0.0)

    def forward(
        self,
        query: torch.Tensor,
        reference_points: torch.Tensor,
        input_flatten: torch.Tensor,
        input_spatial_shapes: torch.Tensor,
        input_level_start_index: torch.Tensor,
        input_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        (N, num_queries, _) = query.size()
        (N, sequence_length, _) = input_flatten.size()
        assert (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1]).sum() == sequence_length

        value = self.value_proj(input_flatten)
        if input_padding_mask is not None:
            value = value.masked_fill(input_padding_mask[..., None], float(0))

        value = value.view(N, sequence_length, self.n_heads, self.d_model // self.n_heads)
        sampling_offsets = self.sampling_offsets(query).view(
            N, num_queries, self.n_heads, self.n_levels, self.n_points, 2
        )
        attention_weights = self.attention_weights(query).view(
            N, num_queries, self.n_heads, self.n_levels * self.n_points
        )
        attention_weights = F.softmax(attention_weights, -1).view(
            N, num_queries, self.n_heads, self.n_levels, self.n_points
        )

        # N, num_queries, n_heads, n_levels, n_points, 2
        if reference_points.shape[-1] == 2:
            offset_normalizer = torch.stack([input_spatial_shapes[..., 1], input_spatial_shapes[..., 0]], -1)
            sampling_locations = (
                reference_points[:, :, None, :, None, :]
                + sampling_offsets / offset_normalizer[None, None, None, :, None, :]
            )
        elif reference_points.shape[-1] == 4:
            sampling_locations = (
                reference_points[:, :, None, :, None, :2]
                + sampling_offsets / self.n_points * reference_points[:, :, None, :, None, 2:] * 0.5
            )
        else:
            raise ValueError(
                f"Last dim of reference_points must be 2 or 4, but get {reference_points.shape[-1]} instead"
            )

        output = self.msda(
            value,
            input_spatial_shapes,
            input_level_start_index,
            sampling_locations,
            attention_weights,
            self.im2col_step,
        )

        output = self.output_proj(output)

        return output


class DeformableTransformerEncoderLayer(nn.Module):
    def __init__(self, d_model: int, d_ffn: int, dropout: float, n_levels: int, n_heads: int, n_points: int) -> None:
        super().__init__()
        self.self_attn = MultiScaleDeformableAttention(d_model, n_levels, n_heads, n_points)
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)

        self.dropout = nn.Dropout(dropout)
        self.activation = nn.ReLU()

    def forward(
        self,
        src: torch.Tensor,
        pos: torch.Tensor,
        reference_points: torch.Tensor,
        spatial_shapes: torch.Tensor,
        level_start_index: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        src2 = self.self_attn(src + pos, reference_points, src, spatial_shapes, level_start_index, mask)
        src = src + self.dropout(src2)
        src = self.norm1(src)

        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout(src2)
        src = self.norm2(src)

        return src


class DeformableTransformerDecoderLayer(nn.Module):
    def __init__(self, d_model: int, d_ffn: int, dropout: float, n_levels: int, n_heads: int, n_points: int) -> None:
        super().__init__()

        # Self attention
        self.self_attn = nn.MultiheadAttention(d_model, n_heads, dropout=dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # Cross attention
        self.cross_attn = MultiScaleDeformableAttention(d_model, n_levels, n_heads, n_points)
        self.norm2 = nn.LayerNorm(d_model)

        # FFN
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm3 = nn.LayerNorm(d_model)

        self.activation = nn.ReLU()
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        tgt: torch.Tensor,
        query_pos: torch.Tensor,
        reference_points: torch.Tensor,
        src: torch.Tensor,
        src_spatial_shapes: torch.Tensor,
        level_start_index: torch.Tensor,
        src_padding_mask: torch.Tensor,
    ) -> torch.Tensor:
        # Self attention
        q = tgt + query_pos
        k = tgt + query_pos

        tgt2 = self.self_attn(q.transpose(0, 1), k.transpose(0, 1), tgt.transpose(0, 1))[0].transpose(0, 1)
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm1(tgt)

        # Cross attention
        tgt2 = self.cross_attn(
            tgt + query_pos, reference_points, src, src_spatial_shapes, level_start_index, src_padding_mask
        )
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm2(tgt)

        # FFN
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm3(tgt)

        return tgt


class DeformableTransformerEncoder(nn.Module):
    def __init__(self, encoder_layer: nn.Module, num_layers: int) -> None:
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)

    @staticmethod
    def get_reference_points(
        spatial_shapes: torch.Tensor, valid_ratios: torch.Tensor, device: torch.device
    ) -> torch.Tensor:
        reference_points_list = []
        for lvl, spatial_shape in enumerate(spatial_shapes):
            H = spatial_shape[0]
            W = spatial_shape[1]
            (ref_y, ref_x) = torch.meshgrid(
                torch.linspace(0.5, H - 0.5, H, dtype=torch.float32, device=device),
                torch.linspace(0.5, W - 0.5, W, dtype=torch.float32, device=device),
                indexing="ij",
            )
            ref_y = ref_y.reshape(-1)[None] / (valid_ratios[:, None, lvl, 1] * H)
            ref_x = ref_x.reshape(-1)[None] / (valid_ratios[:, None, lvl, 0] * W)
            ref = torch.stack((ref_x, ref_y), dim=-1)
            reference_points_list.append(ref)

        reference_points = torch.concat(reference_points_list, dim=1)
        reference_points = reference_points[:, :, None] * valid_ratios[:, None]

        return reference_points

    def forward(
        self,
        src: torch.Tensor,
        spatial_shapes: torch.Tensor,
        level_start_index: torch.Tensor,
        pos: torch.Tensor,
        valid_ratios: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        out = src
        reference_points = self.get_reference_points(spatial_shapes, valid_ratios, device=src.device)
        for layer in self.layers:
            out = layer(out, pos, reference_points, spatial_shapes, level_start_index, mask)

        return out


class DeformableTransformerDecoder(nn.Module):
    def __init__(self, decoder_layer: nn.Module, num_layers: int, return_intermediate: bool) -> None:
        super().__init__()
        self.num_layers = num_layers
        self.layers = _get_clones(decoder_layer, num_layers)
        self.return_intermediate = return_intermediate

        self.box_refine = False
        self.bbox_embed = nn.ModuleList([nn.Identity() for _ in range(self.num_layers)])

    def forward(
        self,
        tgt: torch.Tensor,
        reference_points: torch.Tensor,
        src: torch.Tensor,
        src_spatial_shapes: torch.Tensor,
        src_level_start_index: torch.Tensor,
        query_pos: torch.Tensor,
        src_valid_ratios: torch.Tensor,
        src_padding_mask: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        output = tgt

        intermediate = []
        intermediate_reference_points = []
        for layer, bbox_embed in zip(self.layers, self.bbox_embed):
            if reference_points.shape[-1] == 4:
                reference_points_input = (
                    reference_points[:, :, None] * torch.concat([src_valid_ratios, src_valid_ratios], dim=-1)[:, None]
                )
            else:
                assert reference_points.shape[-1] == 2
                reference_points_input = reference_points[:, :, None] * src_valid_ratios[:, None]

            output = layer(
                output,
                query_pos,
                reference_points_input,
                src,
                src_spatial_shapes,
                src_level_start_index,
                src_padding_mask,
            )

            if self.box_refine is True:
                tmp = bbox_embed(output)
                if reference_points.shape[-1] == 4:
                    new_reference_points = tmp + inverse_sigmoid(reference_points)
                    new_reference_points = new_reference_points.sigmoid()
                else:
                    assert reference_points.shape[-1] == 2
                    new_reference_points = tmp
                    new_reference_points[..., :2] = tmp[..., :2] + inverse_sigmoid(reference_points)
                    new_reference_points = new_reference_points.sigmoid()

                reference_points = new_reference_points.detach()

            if self.return_intermediate is True:
                intermediate.append(output)
                intermediate_reference_points.append(reference_points)

        if self.return_intermediate is True:
            return (torch.stack(intermediate), torch.stack(intermediate_reference_points))

        return (output, reference_points)


class DeformableTransformer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        dim_feedforward: int,
        dropout: float,
        return_intermediate_dec: bool,
        num_feature_levels: int,
        dec_n_points: int,
        enc_n_points: int,
    ) -> None:
        super().__init__()
        self.d_model = d_model
        self.num_heads = num_heads

        encoder_layer = DeformableTransformerEncoderLayer(
            d_model, dim_feedforward, dropout, num_feature_levels, num_heads, enc_n_points
        )
        self.encoder = DeformableTransformerEncoder(encoder_layer, num_encoder_layers)

        decoder_layer = DeformableTransformerDecoderLayer(
            d_model, dim_feedforward, dropout, num_feature_levels, num_heads, dec_n_points
        )
        self.decoder = DeformableTransformerDecoder(decoder_layer, num_decoder_layers, return_intermediate_dec)
        self.level_embed = nn.Parameter(torch.Tensor(num_feature_levels, d_model))
        self.reference_points = nn.Linear(d_model, 2)

        # Weights initialization
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

        for m in self.modules():
            if isinstance(m, MultiScaleDeformableAttention):
                m._reset_parameters()

            nn.init.xavier_uniform_(self.reference_points.weight.data, gain=1.0)
            nn.init.zeros_(self.reference_points.bias.data)

        nn.init.normal_(self.level_embed)

    def get_valid_ratio(self, mask: torch.Tensor) -> torch.Tensor:
        (_, H, W) = mask.size()
        valid_h = torch.sum(~mask[:, :, 0], 1)
        valid_w = torch.sum(~mask[:, 0, :], 1)
        valid_ratio_h = valid_h.float() / H
        valid_ratio_w = valid_w.float() / W
        valid_ratio = torch.stack([valid_ratio_w, valid_ratio_h], dim=-1)

        return valid_ratio

    # pylint: disable=too-many-locals
    def forward(
        self,
        srcs: list[torch.Tensor],
        pos_embeds: list[torch.Tensor],
        query_embed: torch.Tensor,
        masks: list[torch.Tensor],
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # Prepare input for encoder
        src_list = []
        lvl_pos_embed_list = []
        mask_list = []
        spatial_shape_list: list[list[int]] = []  # list[tuple[int, int]] not supported on TorchScript
        for lvl, (src, pos_embed, mask) in enumerate(zip(srcs, pos_embeds, masks)):
            (_, _, H, W) = src.size()
            spatial_shape_list.append([H, W])
            src = src.flatten(2).transpose(1, 2)
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
            mask = mask.flatten(1)
            lvl_pos_embed = pos_embed + self.level_embed[lvl].view(1, 1, -1)
            lvl_pos_embed_list.append(lvl_pos_embed)
            src_list.append(src)
            mask_list.append(mask)

        src_flatten = torch.concat(src_list, dim=1)
        mask_flatten = torch.concat(mask_list, dim=1)
        lvl_pos_embed_flatten = torch.concat(lvl_pos_embed_list, dim=1)
        spatial_shapes = torch.as_tensor(spatial_shape_list, dtype=torch.long, device=src_flatten.device)
        level_start_index = torch.concat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]), dim=0)
        valid_ratios = torch.stack([self.get_valid_ratio(m) for m in masks], dim=1)

        # Encoder
        memory = self.encoder(
            src_flatten, spatial_shapes, level_start_index, lvl_pos_embed_flatten, valid_ratios, mask_flatten
        )

        # Prepare input for decoder
        (B, _, C) = memory.size()
        query_embed, tgt = torch.split(query_embed, C, dim=1)
        query_embed = query_embed.unsqueeze(0).expand(B, -1, -1)
        tgt = tgt.unsqueeze(0).expand(B, -1, -1)
        reference_points = self.reference_points(query_embed).sigmoid()

        # Decoder
        (hs, inter_references) = self.decoder(
            tgt, reference_points, memory, spatial_shapes, level_start_index, query_embed, valid_ratios, mask_flatten
        )

        return (hs, reference_points, inter_references)


# pylint: disable=invalid-name
class Deformable_DETR(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        # Sigmoid based classification (like multi-label networks)
        self.num_classes = self.num_classes - 1

        hidden_dim = 256
        num_heads = 8
        dim_feedforward = 1024
        dec_n_points = 4
        enc_n_points = 4
        dropout: float = self.config.get("dropout", 0.1)
        num_encoder_layers: int = self.config.get("num_encoder_layers", 6)
        num_decoder_layers: int = self.config.get("num_decoder_layers", 6)
        num_queries: int = self.config.get("num_queries", 300)
        box_refine: bool = self.config["box_refine"]
        soft_nms: bool = self.config.get("soft_nms", False)

        self.soft_nms = None
        if soft_nms is True:
            self.soft_nms = SoftNMS()

        self.nms_thresh = 0.5
        self.box_refine = box_refine
        self.hidden_dim = hidden_dim
        input_proj_list = []
        for ch in self.backbone.return_channels:
            input_proj_list.append(
                nn.Sequential(
                    nn.Conv2d(ch, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)),
                    nn.GroupNorm(32, hidden_dim),
                )
            )

        self.input_proj = nn.ModuleList(input_proj_list)

        self.transformer = DeformableTransformer(
            d_model=hidden_dim,
            num_heads=num_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            return_intermediate_dec=True,
            num_feature_levels=len(self.backbone.return_channels),
            dec_n_points=dec_n_points,
            enc_n_points=enc_n_points,
        )

        num_pred = self.transformer.decoder.num_layers

        self.query_embed = nn.Embedding(num_queries, hidden_dim * 2)
        self.pos_enc = PositionEmbeddingSine(hidden_dim // 2, normalize=True)
        self.matcher = HungarianMatcher(cost_class=2, cost_bbox=5, cost_giou=2)

        class_embed = nn.Linear(hidden_dim, self.num_classes)
        bbox_embed = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)
        if self.box_refine is True:
            self.class_embed = _get_clones(class_embed, num_pred)
            self.bbox_embed = _get_clones(bbox_embed, num_pred)
            self.transformer.decoder.box_refine = True
            self.transformer.decoder.bbox_embed = self.bbox_embed
        else:
            self.class_embed = nn.ModuleList([class_embed for _ in range(num_pred)])
            self.bbox_embed = nn.ModuleList([bbox_embed for _ in range(num_pred)])

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

        # Weights initialization
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        for class_embed in self.class_embed:
            class_embed.bias.data = torch.ones(self.num_classes) * bias_value

        for proj in self.input_proj:
            nn.init.xavier_uniform_(proj[0].weight, gain=1)
            nn.init.zeros_(proj[0].bias)

        for bbox_embed in self.bbox_embed:
            nn.init.zeros_(bbox_embed[-2].weight.data)
            nn.init.zeros_(bbox_embed[-2].bias.data)

        nn.init.constant_(self.bbox_embed[0][-2].bias.data[2:], -2.0)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        class_embed = nn.Linear(self.hidden_dim, num_classes)
        num_pred = self.transformer.decoder.num_layers
        if self.box_refine is True:
            self.class_embed = _get_clones(class_embed, num_pred)
        else:
            self.class_embed = nn.ModuleList([class_embed for _ in range(num_pred)])

        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        for class_embed in self.class_embed:
            class_embed.bias.data = torch.ones(self.num_classes) * bias_value

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.class_embed.parameters():
                param.requires_grad = True

    def _get_src_permutation_idx(self, indices: list[torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _class_loss(
        self,
        cls_logits: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[torch.Tensor],
        num_boxes: int,
    ) -> torch.Tensor:
        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)
        target_classes = torch.full(cls_logits.shape[:2], self.num_classes, dtype=torch.int64, device=cls_logits.device)
        target_classes[idx] = target_classes_o

        target_classes_onehot = torch.zeros(
            [cls_logits.shape[0], cls_logits.shape[1], cls_logits.shape[2] + 1],
            dtype=cls_logits.dtype,
            layout=cls_logits.layout,
            device=cls_logits.device,
        )
        target_classes_onehot.scatter_(2, target_classes.unsqueeze(-1), 1)

        target_classes_onehot = target_classes_onehot[:, :, :-1]
        loss = sigmoid_focal_loss(cls_logits, target_classes_onehot, alpha=0.25, gamma=2.0)
        loss_ce = (loss.mean(1).sum() / num_boxes) * cls_logits.shape[1]

        return loss_ce

    def _box_loss(
        self,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[torch.Tensor],
        num_boxes: int,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)
        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none")
        loss_bbox = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(
            box_ops.generalized_box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        )
        loss_giou = loss_giou.sum() / num_boxes

        return (loss_bbox, loss_giou)

    @torch.jit.unused  # type: ignore[misc]
    @torch.compiler.disable()  # type: ignore[misc]
    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
    ) -> dict[str, torch.Tensor]:
        # Compute the average number of target boxes across all nodes, for normalization purposes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=cls_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1).item()

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []
        for idx in range(cls_logits.size(0)):
            indices = self.matcher(cls_logits[idx], box_output[idx], targets)
            loss_ce_i = self._class_loss(cls_logits[idx], targets, indices, num_boxes)
            (loss_bbox_i, loss_giou_i) = self._box_loss(box_output[idx], targets, indices, num_boxes)
            loss_ce_list.append(loss_ce_i)
            loss_bbox_list.append(loss_bbox_i)
            loss_giou_list.append(loss_giou_i)

        loss_ce = torch.stack(loss_ce_list).sum() * 2
        loss_bbox = torch.stack(loss_bbox_list).sum() * 5
        loss_giou = torch.stack(loss_giou_list).sum() * 2
        losses = {
            "labels": loss_ce,
            "boxes": loss_bbox,
            "giou": loss_giou,
        }

        return losses

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_shapes: list[tuple[int, int]]
    ) -> list[dict[str, torch.Tensor]]:
        prob = class_logits.sigmoid()
        (topk_values, topk_indexes) = torch.topk(prob.view(class_logits.shape[0], -1), k=100, dim=1)
        scores = topk_values
        topk_boxes = topk_indexes // class_logits.shape[2]
        labels = topk_indexes % class_logits.shape[2]
        labels += 1  # Background offset

        target_sizes = torch.tensor(image_shapes, device=class_logits.device)

        # Convert to [x0, y0, x1, y1] format
        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")
        boxes = torch.gather(boxes, 1, topk_boxes.unsqueeze(-1).repeat(1, 1, 4))

        # Convert from relative [0, 1] to absolute [0, height] coordinates
        (img_h, img_w) = target_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, l, b in zip(scores, labels, boxes):
            # Non-maximum suppression
            if self.soft_nms is not None:
                # Actually much faster on CPU
                device = b.device
                (soft_scores, keep) = self.soft_nms(b.cpu(), s.cpu(), l.cpu(), score_threshold=0.001)
                keep = keep.to(device)
                s[keep] = soft_scores.to(device)

                b = b[keep]
                s = s[keep]
                l = l[keep]  # noqa: E741

            detections.append(
                {
                    "boxes": b,
                    "scores": s,
                    "labels": l,
                }
            )

        return detections

    # pylint: disable=too-many-locals
    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[list[int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        feature_list = list(features.values())
        mask_list = []
        pos_list = []
        for idx, proj in enumerate(self.input_proj):
            if masks is not None:
                mask_size = feature_list[idx].shape[-2:]
                m = F.interpolate(masks[None].float(), size=mask_size, mode="nearest").to(torch.bool)[0]
            else:
                (B, _, H, W) = feature_list[idx].size()
                m = torch.zeros(B, H, W, dtype=torch.bool, device=x.device)

            feature_list[idx] = proj(feature_list[idx])
            mask_list.append(m)
            pos_list.append(self.pos_enc(feature_list[idx], m))

        (hs, init_reference, inter_references) = self.transformer(
            feature_list, pos_list, self.query_embed.weight, mask_list
        )
        outputs_classes = []
        outputs_coords = []
        for lvl, (class_embed, bbox_embed) in enumerate(zip(self.class_embed, self.bbox_embed)):
            if lvl == 0:
                reference = init_reference
            else:
                reference = inter_references[lvl - 1]

            reference = inverse_sigmoid(reference)
            outputs_class = class_embed(hs[lvl])
            tmp = bbox_embed(hs[lvl])
            if reference.shape[-1] == 4:
                tmp += reference
            else:
                assert reference.shape[-1] == 2
                tmp[..., :2] += reference

            outputs_coord = tmp.sigmoid()
            outputs_classes.append(outputs_class)
            outputs_coords.append(outputs_coord)

        outputs_class = torch.stack(outputs_classes)
        outputs_coord = torch.stack(outputs_coords)

        losses = {}
        detections: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            assert targets is not None, "targets should not be none when in training mode"

            # Convert target boxes and classes
            for idx, target in enumerate(targets):
                boxes = target["boxes"]
                boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
                boxes = boxes / torch.tensor(images.image_sizes[idx][::-1] * 2, dtype=torch.float32, device=x.device)
                targets[idx]["boxes"] = boxes
                targets[idx]["labels"] = target["labels"] - 1  # No background

            losses = self.compute_loss(targets, outputs_class, outputs_coord)

        else:
            detections = self.postprocess_detections(outputs_class[-1], outputs_coord[-1], images.image_sizes)

        return (detections, losses)


registry.register_model_config("deformable_detr", Deformable_DETR, config={"box_refine": False})
registry.register_model_config("deformable_detr_boxref", Deformable_DETR, config={"box_refine": True})

registry.register_weights(
    "deformable_detr_boxref_coco_convnext_v2_tiny_imagenet21k",
    {
        "url": (
            "https://huggingface.co/birder-project/"
            "deformable_detr_boxref_coco_convnext_v2_tiny_imagenet21k/resolve/main"
        ),
        "description": (
            "Deformable DETR box refinement with a ConvNeXt v2 tiny backbone pretrained on ImageNet 21K, "
            "detection model trained on the COCO dataset"
        ),
        "resolution": (640, 640),
        "formats": {
            "pt": {
                "file_size": 152.8,
                "sha256": "bd3ee455a98569e19b36de550784995fffd78fd27ac836fc94c18f194b2df580",
            }
        },
        "net": {"network": "deformable_detr_boxref", "tag": "coco"},
        "backbone": {"network": "convnext_v2_tiny", "tag": "imagenet21k"},
    },
)



"""
MobileOne, adapted from
https://github.com/apple/ml-mobileone/blob/main/mobileone.py

Paper "MobileOne: An Improved One millisecond Mobile Backbone",
https://arxiv.org/abs/2206.04040
"""

# Reference license: Apple MIT License

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone


class MobileOneBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        padding: int,
        groups: int,
        use_se: bool,
        use_act: bool,
        use_scale_branch: bool,
        num_conv_branches: int,
        reparameterized: bool,
        activation_layer: Callable[..., nn.Module] = nn.ReLU,
    ) -> None:
        super().__init__()
        self.id_tensor = None
        self.reparameterized = reparameterized
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.groups = groups
        self.num_conv_branches = num_conv_branches

        if use_se is True:
            self.se = SqueezeExcitation(out_channels, out_channels // 16)
        else:
            self.se = nn.Identity()

        if use_act is True:
            self.activation = activation_layer()
        else:
            self.activation = nn.Identity()

        if reparameterized is True:
            self.reparam_conv = nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=groups,
                bias=True,
            )
        else:
            self.reparam_conv = None

            # Re-parameterizable skip connection
            self.rbr_skip = None
            if out_channels == in_channels and stride == 1:
                self.rbr_skip = nn.BatchNorm2d(num_features=in_channels)

            # Re-parameterizable conv branches
            self.rbr_conv = nn.ModuleList()
            for _ in range(self.num_conv_branches):
                mod_list = nn.Sequential()
                mod_list.add_module(
                    "conv",
                    nn.Conv2d(
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=padding,
                        groups=self.groups,
                        bias=False,
                    ),
                )
                mod_list.add_module("bn", nn.BatchNorm2d(num_features=out_channels))
                self.rbr_conv.append(mod_list)

            # Re-parameterizable scale branch
            self.rbr_scale = None
            if kernel_size > 1 and use_scale_branch is True:
                self.rbr_scale = nn.Sequential()
                self.rbr_scale.add_module(
                    "conv",
                    nn.Conv2d(
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=(1, 1),
                        stride=stride,
                        padding=(0, 0),
                        groups=self.groups,
                        bias=False,
                    ),
                )
                self.rbr_scale.add_module("bn", nn.BatchNorm2d(num_features=out_channels))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # Reparameterized forward pass
        if self.reparam_conv is not None:
            return self.activation(self.se(self.reparam_conv(x)))

        # Multi-branched train-time forward pass
        identity_out = 0
        if self.rbr_skip is not None:
            identity_out = self.rbr_skip(x)

        scale_out = 0
        if self.rbr_scale is not None:
            scale_out = self.rbr_scale(x)

        out = scale_out + identity_out
        for conv in self.rbr_conv:
            out += conv(x)

        return self.activation(self.se(out))

    def reparameterize(self) -> None:
        """
        Following works like "RepVGG: Making VGG-style ConvNets Great Again" - https://arxiv.org/pdf/2101.03697.pdf.
        We re-parameterize multi-branched architecture used at training time
        to obtain a plain CNN-like structure for inference.
        """

        if self.reparameterized is True:
            return

        (kernel, bias) = self._get_kernel_bias()
        self.reparam_conv = nn.Conv2d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            groups=self.groups,
            bias=True,
        )
        self.reparam_conv.weight.data = kernel
        self.reparam_conv.bias.data = bias

        # Delete un-used branches
        for param in self.parameters():
            param.detach_()

        del self.rbr_conv
        del self.rbr_scale
        if hasattr(self, "rbr_skip") is True:
            del self.rbr_skip

        self.reparameterized = True

    def _get_kernel_bias(self) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Method to obtain re-parameterized kernel and bias.
        Reference: https://github.com/DingXiaoH/RepVGG/blob/main/repvgg.py#L83
        """

        # Get weights and bias of scale branch
        kernel_scale = 0
        bias_scale = 0
        if self.rbr_scale is not None:
            (kernel_scale, bias_scale) = self._fuse_bn_tensor(self.rbr_scale)
            pad = self.kernel_size // 2
            kernel_scale = F.pad(kernel_scale, [pad, pad, pad, pad])

        # Get weights and bias of skip branch
        kernel_identity = 0
        bias_identity = 0
        if self.rbr_skip is not None:
            (kernel_identity, bias_identity) = self._fuse_bn_tensor(self.rbr_skip)

        # Get weights and bias of conv branches
        kernel_conv = 0
        bias_conv = 0
        for ix in range(self.num_conv_branches):
            (_kernel, _bias) = self._fuse_bn_tensor(self.rbr_conv[ix])
            kernel_conv += _kernel
            bias_conv += _bias

        kernel_final = kernel_conv + kernel_scale + kernel_identity
        bias_final = bias_conv + bias_scale + bias_identity

        return (kernel_final, bias_final)

    def _fuse_bn_tensor(self, branch: nn.Module) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Method to fuse batchnorm layer with preceding convolution layer.
        Reference: https://github.com/DingXiaoH/RepVGG/blob/main/repvgg.py#L95
        """

        if isinstance(branch, nn.Sequential):
            kernel_value = branch.conv.weight
            running_mean = branch.bn.running_mean
            running_var = branch.bn.running_var
            gamma = branch.bn.weight
            beta = branch.bn.bias
            eps = branch.bn.eps

        else:
            assert isinstance(branch, nn.BatchNorm2d)
            input_dim = self.in_channels // self.groups
            kernel_value = torch.zeros(
                (self.in_channels, input_dim, self.kernel_size, self.kernel_size),
                dtype=branch.weight.dtype,
                device=branch.weight.device,
            )
            for i in range(self.in_channels):
                kernel_value[i, i % input_dim, self.kernel_size // 2, self.kernel_size // 2] = 1

            self.id_tensor = kernel_value

            running_mean = branch.running_mean
            running_var = branch.running_var
            gamma = branch.weight
            beta = branch.bias
            eps = branch.eps

        std = (running_var + eps).sqrt()
        t = (gamma / std).reshape(-1, 1, 1, 1)

        return (kernel_value * t, beta - running_mean * gamma / std)


class MobileOneStage(nn.Sequential):
    def __init__(
        self,
        in_planes: int,
        planes: int,
        num_blocks: int,
        num_se_blocks: int,
        num_conv_branches: int,
        reparameterized: bool,
    ) -> None:
        super().__init__()
        strides = [2] + [1] * (num_blocks - 1)
        for idx, stride in enumerate(strides):
            use_se = False
            if idx >= (num_blocks - num_se_blocks):
                use_se = True

            # Depthwise conv
            self.append(
                MobileOneBlock(
                    in_channels=in_planes,
                    out_channels=in_planes,
                    kernel_size=3,
                    stride=stride,
                    padding=1,
                    groups=in_planes,
                    use_se=use_se,
                    use_act=True,
                    use_scale_branch=True,
                    num_conv_branches=num_conv_branches,
                    reparameterized=reparameterized,
                )
            )
            # Pointwise conv
            self.append(
                MobileOneBlock(
                    in_channels=in_planes,
                    out_channels=planes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    groups=1,
                    use_se=use_se,
                    use_act=True,
                    use_scale_branch=True,
                    num_conv_branches=num_conv_branches,
                    reparameterized=reparameterized,
                )
            )
            in_planes = planes


class MobileOne(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        self.reparameterized = False
        num_blocks_per_stage = [2, 8, 10, 1]
        widths = [64, 128, 256, 512]
        width_multipliers: list[float] = self.config["width_multipliers"]
        num_conv_branches: int = self.config["num_conv_branches"]
        num_se_blocks: list[int] = self.config["num_se_blocks"]

        in_planes = min(64, int(widths[0] * width_multipliers[0]))
        self.stem = MobileOneBlock(
            in_channels=self.input_channels,
            out_channels=in_planes,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=1,
            use_se=False,
            use_act=True,
            use_scale_branch=True,
            num_conv_branches=1,
            reparameterized=self.reparameterized,
        )

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx in range(len(num_blocks_per_stage)):  # pylint: disable=consider-using-enumerate
            out_planes = int(widths[idx] * width_multipliers[idx])
            stages[f"stage{idx+1}"] = MobileOneStage(
                in_planes,
                out_planes,
                num_blocks_per_stage[idx],
                num_se_blocks=num_se_blocks[idx],
                num_conv_branches=num_conv_branches,
                reparameterized=self.reparameterized,
            )
            return_channels.append(out_planes)
            in_planes = out_planes

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = int(widths[-1] * width_multipliers[3])
        self.classifier = self.create_classifier()

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def reparameterize_model(self) -> None:
        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config(
    "mobileone_s0",
    MobileOne,
    config={"width_multipliers": [0.75, 1.0, 1.0, 2.0], "num_conv_branches": 4, "num_se_blocks": [0, 0, 0, 0]},
)
registry.register_model_config(
    "mobileone_s1",
    MobileOne,
    config={"width_multipliers": [1.5, 1.5, 2.0, 2.5], "num_conv_branches": 1, "num_se_blocks": [0, 0, 0, 0]},
)
registry.register_model_config(
    "mobileone_s2",
    MobileOne,
    config={"width_multipliers": [1.5, 2.0, 2.5, 4.0], "num_conv_branches": 1, "num_se_blocks": [0, 0, 0, 0]},
)
registry.register_model_config(
    "mobileone_s3",
    MobileOne,
    config={"width_multipliers": [2.0, 2.5, 3.0, 4.0], "num_conv_branches": 1, "num_se_blocks": [0, 0, 0, 0]},
)
registry.register_model_config(
    "mobileone_s4",
    MobileOne,
    config={"width_multipliers": [3.0, 3.5, 3.5, 4.0], "num_conv_branches": 1, "num_se_blocks": [0, 0, 5, 1]},
)

registry.register_weights(
    "mobileone_s0_il-common",
    {
        "description": "MobileOne S0 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 18.5,
                "sha256": "9643db45433e92078a7c29fccec2dc8d787d638ce42e4565f3df330f4c761b3f",
            }
        },
        "net": {"network": "mobileone_s0", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobileone_s0_il-common_reparameterized",
    {
        "description": "MobileOne S0 (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 5.5,
                "sha256": "c400cd634c02134de7bdbc68c98f92aa167c0d495bd3ed9bd2613fde7dc3198f",
            }
        },
        "net": {"network": "mobileone_s0", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)
registry.register_weights(
    "mobileone_s1_il-common",
    {
        "description": "MobileOne S1 model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 15.8,
                "sha256": "10b47b9b90cdad7f19130b75440a01d8d7b625882903d752927e6acabaf612ee",
            }
        },
        "net": {"network": "mobileone_s1", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobileone_s1_il-common_reparameterized",
    {
        "description": "MobileOne S1 (reparameterized) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 15.1,
                "sha256": "4455b027f1e6332d93f237c34c440b49e3f9c8b6fe711765e20ae57f14aa2738",
            }
        },
        "net": {"network": "mobileone_s1", "tag": "il-common_reparameterized", "reparameterized": True},
    },
)
